
-- | Color spaces

module Data.ColorSpaces 
  ( HSL(..)
  , HSV(..) 
  , RGB(..)
  , ColorSpace(..)
  , gammaCorrection
--  , hsl_to_rgb
--  , hsv_to_rgb
--  , rgb_to_hsl
--  , rgb_to_hsv
--  , clamp
--  , scaleinto
  ) where

-------------------------------------------------------

newtype RGB = RGB { unRGB :: (Float,Float,Float) } deriving (Read,Show)
newtype HSL = HSL { unHSL :: (Float,Float,Float) } deriving (Read,Show)
newtype HSV = HSV { unHSV :: (Float,Float,Float) } deriving (Read,Show)

class ColorSpace a where
  toRGB   :: a -> RGB
  fromRGB :: RGB -> a

instance ColorSpace RGB where
  toRGB   = id
  fromRGB = id

instance ColorSpace HSV where
  toRGB   = hsv_to_rgb
  fromRGB = rgb_to_hsv
  
instance ColorSpace HSL where
  toRGB   = hsl_to_rgb
  fromRGB = rgb_to_hsl

-------------------------------------------------------

{-# INLINE ffrac #-}
ffrac :: Float -> Float
ffrac x = x - fromIntegral (floor x)

{-# INLINE scaleinto #-}
{-# SPECIALIZE scaleinto :: (Float,Float) -> Float -> Float #-}
{-# SPECIALIZE scaleinto :: (Double,Double) -> Double -> Double #-}
scaleinto :: Floating a => (a,a) -> a -> a
scaleinto (a,b) x = x*(b-a)+a

{-# INLINE clamp #-}
clamp :: Float -> Float
clamp a = max (min a 1) 0

-------------------------------------------------------

gammaCorrection :: Float -> RGB -> RGB
gammaCorrection gamma (RGB (r,g,b)) = RGB (f r, f g, f b) where
  f x = exp $ gamma * (log (x + 0.00000000001))
 
-------------------------------------------------------

oneThird  = 1.0/3.0 :: Float
oneSixth  = 1.0/6.0 :: Float
twoThirds = 2.0/3.0 :: Float

-- source: http://en.wikipedia.org/wiki/HSL_color_space
hsl_to_rgb :: HSL -> RGB
hsl_to_rgb (HSL (h',s',l')) = RGB (r,g,b) where
  h = ffrac h'
  s = clamp s'
  l = clamp l'
  q = if l<0.5
    then l*(1+s)
    else l+s-(l*s)
  p = 2*l-q
  tl = map ( ffrac . (+h) ) [oneThird,0,-oneThird]
  [r,g,b] = map f tl
  f x = -- trace (show x) $ 
    if x < oneSixth then p + (q-p)*6*x 
      else if x < 0.5 then q
        else if x < twoThirds then p + (q-p)*6*(twoThirds-x)
          else p
          
-- source: http://en.wikipedia.org/wiki/HSV_color_space
hsv_to_rgb :: HSV -> RGB
hsv_to_rgb (HSV (h',s',v')) = RGB rgb where
  h = ffrac h'
  s = clamp s'
  v = clamp v'
  (i,f) = properFraction (h*6)
  p = v*(1-s)
  q = v*(1-f*s)
  t = v*(1-(1-f)*s)
  rgb = case i of
    0 -> (v,t,p)
    1 -> (q,v,p)
    2 -> (p,v,t)
    3 -> (p,q,v)
    4 -> (t,p,v)
    5 -> (v,p,q)
    6 -> (v,t,p) -- hmm hmm, ugy tunik valamiert ez is elofordul?
    
-------------------------------------------------------

hue (mn,mx) (r,g,b) 
  | mx == mn  = 0
  | mx == r   = ffrac $             oneSixth*(g-b)/diff 
  | mx == g   = ffrac $ oneThird  + oneSixth*(b-r)/diff 
  | mx == b   = ffrac $ twoThirds + oneSixth*(r-g)/diff 
  where diff = mx - mn
  
hsl_sat l mn mx 
  | mx == mn   = 0
  | l  <= 0.5  = 0.5 * (mx-mn) /    l
  | l  >  0.5  = 0.5 * (mx-mn) / (1-l)
  
rgb_to_hsl :: RGB -> HSL
rgb_to_hsl (RGB rgb@(r,g,b)) = HSL (h,s,l) where
  rgblist = [r,g,b]
  mn = minimum rgblist
  mx = maximum rgblist
  h = hue (mn,mx) rgb
  l = 0.5 * (mn+mx)
  s = hsl_sat l mn mx

rgb_to_hsv :: RGB -> HSV
rgb_to_hsv (RGB rgb@(r,g,b)) = HSV (h,s,mx) where
  rgblist = [r,g,b]
  mn = minimum rgblist
  mx = maximum rgblist
  h = hue (mn,mx) rgb
  s = if mx>0 then (mx-mn) / mx else 0
    
-------------------------------------------------------
    